@@ -26,8 +26,7 @@
 #include <stdlib.h>
 #include <memory.h>
 #include <math.h>
-#include <glib/gi18n.h>
-#include <libxklavier/xklavier.h>
+#include <glib/gi18n-lib.h>
 
 #include <gkbd-keyboard-drawing.h>
 #include <gkbd-keyboard-drawing-marshal.h>
@@ -39,6 +38,8 @@
 
 #define GTK_RESPONSE_PRINT 2
 
+#define CAIRO_LINE_WIDTH 1.0
+
 enum {
 	BAD_KEYCODE = 0,
 	NUM_SIGNALS
@@ -46,6 +47,20 @@ enum {
 
 static guint gkbd_keyboard_drawing_signals[NUM_SIGNALS] = { 0 };
 
+static GkbdKeyboardDrawingGroupLevel defaultGroupsLevels[] = {
+	{0, 1},
+	{0, 3},
+	{0, 0},
+	{0, 2}
+};
+
+static GkbdKeyboardDrawingGroupLevel *pGroupsLevels[] = {
+	defaultGroupsLevels,
+	defaultGroupsLevels + 1,
+	defaultGroupsLevels + 2,
+	defaultGroupsLevels + 3
+};
+
 static void gkbd_keyboard_drawing_set_mods (GkbdKeyboardDrawing * drawing,
 					    guint mods);
 
@@ -282,8 +297,10 @@ rounded_polygon (cairo_t * cr,
 
 	if (filled)
 		cairo_fill (cr);
-	else
+	else {
+		cairo_set_line_width (cr, CAIRO_LINE_WIDTH);
 		cairo_stroke (cr);
+	}
 }
 
 static void
@@ -360,23 +377,25 @@ curve_rectangle (cairo_t * cr,
 static void
 draw_curve_rectangle (cairo_t * cr,
 		      gboolean filled,
-		      GdkColor * fill_color,
+		      GdkColor * color,
 		      gint x, gint y, gint width, gint height, gint radius)
 {
 	curve_rectangle (cr, x, y, width, height, radius);
 
-	gdk_cairo_set_source_color (cr, fill_color);
+	gdk_cairo_set_source_color (cr, color);
 
 	if (filled)
 		cairo_fill (cr);
-	else
+	else {
+		cairo_set_line_width (cr, CAIRO_LINE_WIDTH);
 		cairo_stroke (cr);
+	}
 }
 
 /* x, y, width, height are in the xkb coordinate system */
 static void
 draw_rectangle (GkbdKeyboardDrawingRenderContext * context,
-		GdkColor * fill_color,
+		GdkColor * color,
 		gint angle,
 		gint xkb_x, gint xkb_y, gint xkb_width, gint xkb_height,
 		gint radius)
@@ -385,10 +404,10 @@ draw_rectangle (GkbdKeyboardDrawingRenderContext * context,
 		gint x, y, width, height;
 		gboolean filled;
 
-		if (fill_color) {
+		if (color) {
 			filled = TRUE;
 		} else {
-			fill_color = context->dark_color;
+			color = context->dark_color;
 			filled = FALSE;
 		}
 
@@ -399,7 +418,7 @@ draw_rectangle (GkbdKeyboardDrawingRenderContext * context,
 		height =
 		    xkb_to_pixmap_coord (context, xkb_y + xkb_height) - y;
 
-		draw_curve_rectangle (context->cr, filled, fill_color,
+		draw_curve_rectangle (context->cr, filled, color,
 				      x, y, width, height,
 				      xkb_to_pixmap_double (context,
 							    radius));
@@ -423,8 +442,7 @@ draw_rectangle (GkbdKeyboardDrawingRenderContext * context,
 		points[3].y = y;
 
 		/* the points we've calculated are relative to 0,0 */
-		draw_polygon (context, fill_color, 0, 0, points, 4,
-			      radius);
+		draw_polygon (context, color, 0, 0, points, 4, radius);
 	}
 }
 
@@ -434,23 +452,25 @@ draw_outline (GkbdKeyboardDrawingRenderContext * context,
 	      GdkColor * color, gint angle, gint origin_x, gint origin_y)
 {
 #ifdef KBDRAW_DEBUG
-	printf (" num_points in %p: %d\n", outline, outline->num_points);
+	printf ("origin: %d, %d, num_points in %p: %d\n", origin_x,
+		origin_y, outline, outline->num_points);
 #endif
 
 	if (outline->num_points == 1) {
+#ifdef KBDRAW_DEBUG
+		printf
+		    ("1 point (rectangle): width, height: %d, %d, radius %d\n",
+		     outline->points[0].x, outline->points[0].y,
+		     outline->corner_radius);
+#endif
 		if (color)
 			draw_rectangle (context, color, angle, origin_x,
 					origin_y, outline->points[0].x,
 					outline->points[0].y,
 					outline->corner_radius);
 
-#ifdef KBDRAW_DEBUG
-		printf ("pointsxy:%d %d %d\n", outline->points[0].x,
-			outline->points[0].y, outline->corner_radius);
-#endif
-
-		draw_rectangle (context, NULL, angle, origin_x, origin_y,
-				outline->points[0].x,
+		draw_rectangle (context, NULL, angle, origin_x,
+				origin_y, outline->points[0].x,
 				outline->points[0].y,
 				outline->corner_radius);
 	} else if (outline->num_points == 2) {
@@ -460,23 +480,31 @@ draw_outline (GkbdKeyboardDrawingRenderContext * context,
 				   origin_x + outline->points[0].x,
 				   origin_y + outline->points[0].y,
 				   angle, &rotated_x0, &rotated_y0);
+#ifdef KBDRAW_DEBUG
+		printf
+		    ("2 points (rectangle): from %d, %d, width, height: %d, %d, radius %d\n",
+		     rotated_x0, rotated_y0, outline->points[1].x,
+		     outline->points[1].y, outline->corner_radius);
+#endif
 		if (color)
 			draw_rectangle (context, color, angle, rotated_x0,
 					rotated_y0, outline->points[1].x,
 					outline->points[1].y,
 					outline->corner_radius);
-
 		draw_rectangle (context, NULL, angle, rotated_x0,
 				rotated_y0, outline->points[1].x,
 				outline->points[1].y,
 				outline->corner_radius);
 	} else {
+#ifdef KBDRAW_DEBUG
+		printf ("multiple points (%d) from %d %d, radius %d\n",
+			outline->num_points, origin_x, origin_y,
+			outline->corner_radius);
+#endif
 		if (color)
 			draw_polygon (context, color, origin_x, origin_y,
-				      outline->points,
-				      outline->num_points,
+				      outline->points, outline->num_points,
 				      outline->corner_radius);
-
 		draw_polygon (context, NULL, origin_x, origin_y,
 			      outline->points, outline->num_points,
 			      outline->corner_radius);
@@ -494,46 +522,46 @@ parse_xkb_color_spec (gchar * colorspec, GdkColor * color)
 		color->green = 0;
 		color->blue = 0;
 	} else if (g_ascii_strcasecmp (colorspec, "white") == 0) {
-		color->red = 65535;
-		color->green = 65535;
-		color->blue = 65535;
+		color->red = 0xFFFF;
+		color->green = 0xFFFF;
+		color->blue = 0xFFFF;
 	} else if (g_ascii_strncasecmp (colorspec, "grey", 4) == 0 ||
 		   g_ascii_strncasecmp (colorspec, "gray", 4) == 0) {
 		level = strtol (colorspec + 4, NULL, 10);
 
-		color->red = 65535 - 65535 * level / 100;
-		color->green = 65535 - 65535 * level / 100;
-		color->blue = 65535 - 65535 * level / 100;
+		color->red = 0xFFFF - 0xFFFF * level / 100;
+		color->green = 0xFFFF - 0xFFFF * level / 100;
+		color->blue = 0xFFFF - 0xFFFF * level / 100;
 	} else if (g_ascii_strcasecmp (colorspec, "red") == 0) {
-		color->red = 65535;
+		color->red = 0xFFFF;
 		color->green = 0;
 		color->blue = 0;
 	} else if (g_ascii_strcasecmp (colorspec, "green") == 0) {
 		color->red = 0;
-		color->green = 65535;
+		color->green = 0xFFFF;
 		color->blue = 0;
 	} else if (g_ascii_strcasecmp (colorspec, "blue") == 0) {
 		color->red = 0;
 		color->green = 0;
-		color->blue = 65535;
+		color->blue = 0xFFFF;
 	} else if (g_ascii_strncasecmp (colorspec, "red", 3) == 0) {
 		level = strtol (colorspec + 3, NULL, 10);
 
-		color->red = 65535 * level / 100;
+		color->red = 0xFFFF * level / 100;
 		color->green = 0;
 		color->blue = 0;
 	} else if (g_ascii_strncasecmp (colorspec, "green", 5) == 0) {
 		level = strtol (colorspec + 5, NULL, 10);
 
 		color->red = 0;
-		color->green = 65535 * level / 100;;
+		color->green = 0xFFFF * level / 100;;
 		color->blue = 0;
 	} else if (g_ascii_strncasecmp (colorspec, "blue", 4) == 0) {
 		level = strtol (colorspec + 4, NULL, 10);
 
 		color->red = 0;
 		color->green = 0;
-		color->blue = 65535 * level / 100;
+		color->blue = 0xFFFF * level / 100;
 	} else
 		return FALSE;
 
@@ -611,14 +639,15 @@ find_keycode (GkbdKeyboardDrawing * drawing, gchar * key_name)
 }
 
 static void
-set_markup(GkbdKeyboardDrawingRenderContext * context,
-		     gchar *txt)
+set_markup (GkbdKeyboardDrawingRenderContext * context, gchar * txt)
 {
 	PangoLayout *layout = context->layout;
-	txt = strcmp("<", txt) ? txt : "&lt;";
-	txt = strcmp("&", txt) ? txt : "&amp;";
+	txt = strcmp ("<", txt) ? txt : "&lt;";
+	txt = strcmp ("&", txt) ? txt : "&amp;";
 	if (g_utf8_strlen (txt, -1) > 1) {
-		gchar* buf = g_strdup_printf ("<span size=\"x-small\">%s</span>", txt);
+		gchar *buf =
+		    g_strdup_printf ("<span size=\"x-small\">%s</span>",
+				     txt);
 		pango_layout_set_markup (layout, buf, -1);
 		g_free (buf);
 	} else {
@@ -634,128 +663,132 @@ set_key_label_in_layout (GkbdKeyboardDrawingRenderContext * context,
 	gunichar uc;
 
 	switch (keyval) {
+	case GDK_KEY_Caps_Lock:
+		set_markup (context, "Caps\nLock");
+		break;
+
 	case GDK_KEY_Scroll_Lock:
 		set_markup (context, "Scroll\nLock");
 		break;
 
 	case GDK_KEY_space:
-		set_markup(context, "");
+		set_markup (context, "");
 		break;
 
 	case GDK_KEY_Sys_Req:
-		set_markup(context, "Sys Rq");
+		set_markup (context, "Sys Rq");
 		break;
 
 	case GDK_KEY_Page_Up:
-		set_markup(context, "Page\nUp");
+		set_markup (context, "Page\nUp");
 		break;
 
 	case GDK_KEY_Page_Down:
-		set_markup(context, "Page\nDown");
+		set_markup (context, "Page\nDown");
 		break;
 
 	case GDK_KEY_Num_Lock:
-		set_markup(context, "Num\nLock");
+		set_markup (context, "Num\nLock");
 		break;
 
 	case GDK_KEY_KP_Page_Up:
-		set_markup(context, "Pg Up");
+		set_markup (context, "Pg Up");
 		break;
 
 	case GDK_KEY_KP_Page_Down:
-		set_markup(context, "Pg Dn");
+		set_markup (context, "Pg Dn");
 		break;
 
 	case GDK_KEY_KP_Home:
-		set_markup(context, "Home");
+		set_markup (context, "Home");
 		break;
 
 	case GDK_KEY_KP_Left:
-		set_markup(context, "Left");
+		set_markup (context, "Left");
 		break;
 
 	case GDK_KEY_KP_End:
-		set_markup(context, "End");
+		set_markup (context, "End");
 		break;
 
 	case GDK_KEY_KP_Up:
-		set_markup(context, "Up");
+		set_markup (context, "Up");
 		break;
 
 	case GDK_KEY_KP_Begin:
-		set_markup(context, "Begin");
+		set_markup (context, "Begin");
 		break;
 
 	case GDK_KEY_KP_Right:
-		set_markup(context, "Right");
+		set_markup (context, "Right");
 		break;
 
 	case GDK_KEY_KP_Enter:
-		set_markup(context, "Enter");
+		set_markup (context, "Enter");
 		break;
 
 	case GDK_KEY_KP_Down:
-		set_markup(context, "Down");
+		set_markup (context, "Down");
 		break;
 
 	case GDK_KEY_KP_Insert:
-		set_markup(context, "Ins");
+		set_markup (context, "Ins");
 		break;
 
 	case GDK_KEY_KP_Delete:
-		set_markup(context, "Del");
+		set_markup (context, "Del");
 		break;
 
 	case GDK_KEY_dead_grave:
-		set_markup(context, "ˋ");
+		set_markup (context, "ˋ");
 		break;
 
 	case GDK_KEY_dead_acute:
-		set_markup(context, "ˊ");
+		set_markup (context, "ˊ");
 		break;
 
 	case GDK_KEY_dead_circumflex:
-		set_markup(context, "ˆ");
+		set_markup (context, "ˆ");
 		break;
 
 	case GDK_KEY_dead_tilde:
-		set_markup(context, "~");
+		set_markup (context, "~");
 		break;
 
 	case GDK_KEY_dead_macron:
-		set_markup(context, "ˉ");
+		set_markup (context, "ˉ");
 		break;
 
 	case GDK_KEY_dead_breve:
-		set_markup(context, "˘");
+		set_markup (context, "˘");
 		break;
 
 	case GDK_KEY_dead_abovedot:
-		set_markup(context, "˙");
+		set_markup (context, "˙");
 		break;
 
 	case GDK_KEY_dead_diaeresis:
-		set_markup(context, "¨");
+		set_markup (context, "¨");
 		break;
 
 	case GDK_KEY_dead_abovering:
-		set_markup(context, "˚");
+		set_markup (context, "˚");
 		break;
 
 	case GDK_KEY_dead_doubleacute:
-		set_markup(context, "˝");
+		set_markup (context, "˝");
 		break;
 
 	case GDK_KEY_dead_caron:
-		set_markup(context, "ˇ");
+		set_markup (context, "ˇ");
 		break;
 
 	case GDK_KEY_dead_cedilla:
-		set_markup(context, "¸");
+		set_markup (context, "¸");
 		break;
 
 	case GDK_KEY_dead_ogonek:
-		set_markup(context, "˛");
+		set_markup (context, "˛");
 		break;
 
 		/* case GDK_KEY_dead_iota:
@@ -763,33 +796,36 @@ set_key_label_in_layout (GkbdKeyboardDrawingRenderContext * context,
 		 * case GDK_KEY_dead_semivoiced_sound: */
 
 	case GDK_KEY_dead_belowdot:
-		set_markup(context, " ̣");
+		set_markup (context, " ̣");
 		break;
 
 	case GDK_KEY_horizconnector:
-		set_markup(context, "horiz\nconn");
+		set_markup (context, "horiz\nconn");
 		break;
 
 	case GDK_KEY_Mode_switch:
-		set_markup(context, "AltGr");
+		set_markup (context, "AltGr");
 		break;
 
 	case GDK_KEY_Multi_key:
-		set_markup(context, "Compose");
+		set_markup (context, "Compose");
+		break;
+
+	case GDK_KEY_VoidSymbol:
+		set_markup (context, "");
 		break;
 
 	default:
 		uc = gdk_keyval_to_unicode (keyval);
 		if (uc != 0 && g_unichar_isgraph (uc)) {
 			buf[g_unichar_to_utf8 (uc, buf)] = '\0';
-			set_markup(context, buf);
+			set_markup (context, buf);
 		} else {
 			gchar *name = gdk_keyval_name (keyval);
 			if (name) {
-				set_markup(context, name);
-			}
-			else
-				set_markup(context, "");
+				set_markup (context, name);
+			} else
+				set_markup (context, "");
 		}
 	}
 }
@@ -798,7 +834,7 @@ set_key_label_in_layout (GkbdKeyboardDrawingRenderContext * context,
 static void
 draw_pango_layout (GkbdKeyboardDrawingRenderContext * context,
 		   GkbdKeyboardDrawing * drawing,
-		   gint angle, gint x, gint y)
+		   gint angle, gint x, gint y, gboolean is_pressed)
 {
 	PangoLayout *layout = context->layout;
 	GdkColor *color;
@@ -806,6 +842,11 @@ draw_pango_layout (GkbdKeyboardDrawingRenderContext * context,
 	gint x_off, y_off;
 	gint i;
 
+	if (is_pressed)
+		color =
+		    &gtk_widget_get_style (GTK_WIDGET (drawing))->text
+		    [GTK_STATE_SELECTED];
+	else
 	color =
 	    drawing->colors + (drawing->xkb->geom->label_color -
 			       drawing->xkb->geom->colors);
@@ -860,7 +901,8 @@ draw_key_label_helper (GkbdKeyboardDrawingRenderContext * context,
 		       gint angle,
 		       GkbdKeyboardDrawingGroupLevelPosition glp,
 		       gint x,
-		       gint y, gint width, gint height, gint padding)
+		       gint y, gint width, gint height, gint padding,
+		       gboolean is_pressed)
 {
 	gint label_x, label_y, label_max_width, ycell;
 
@@ -918,7 +960,8 @@ draw_key_label_helper (GkbdKeyboardDrawingRenderContext * context,
 	cairo_rectangle (context->cr, x + padding / 2, y + padding / 2,
 			 width - padding, height - padding);
 	cairo_clip (context->cr);
-	draw_pango_layout (context, drawing, angle, label_x, label_y);
+	draw_pango_layout (context, drawing, angle, label_x, label_y,
+			   is_pressed);
 	cairo_restore (context->cr);
 }
 
@@ -928,7 +971,8 @@ draw_key_label (GkbdKeyboardDrawingRenderContext * context,
 		guint keycode,
 		gint angle,
 		gint xkb_origin_x,
-		gint xkb_origin_y, gint xkb_width, gint xkb_height)
+		gint xkb_origin_y, gint xkb_width, gint xkb_height,
+		gboolean is_pressed)
 {
 	gint x, y, width, height;
 	gint padding;
@@ -978,7 +1022,8 @@ draw_key_label (GkbdKeyboardDrawingRenderContext * context,
 				draw_key_label_helper (context, drawing,
 						       keysym, angle, glp,
 						       x, y, width, height,
-						       padding);
+						       padding,
+						       is_pressed);
 				/* reverse y order */
 			}
 		} else {
@@ -989,12 +1034,38 @@ draw_key_label (GkbdKeyboardDrawingRenderContext * context,
 
 			draw_key_label_helper (context, drawing, keysym,
 					       angle, glp, x, y, width,
-					       height, padding);
+					       height, padding,
+					       is_pressed);
 			/* reverse y order */
 		}
 	}
 }
 
+/**
+   The x offset is calculated for complex shapes. It is the rightmost of the vertical lines in the outline
+ */
+static gint
+calc_shape_origin_offset_x (XkbOutlineRec * outline)
+{
+	gint rv = 0;
+	gint i;
+	XkbPointPtr point = outline->points;
+	if (outline->num_points < 3)
+		return 0;
+	for (i = outline->num_points; --i > 0;) {
+		gint x1 = point->x;
+		gint y1 = point++->y;
+		gint x2 = point->x;
+		gint y2 = point->y;
+
+		/*vertical, bottom to top (clock-wise), on the left */
+		if ((x1 == x2) && (y1 > y2) && (x1 > rv)) {
+			rv = x1;
+		}
+	}
+	return rv;
+}
+
 /* groups are from 0-3 */
 static void
 draw_key (GkbdKeyboardDrawingRenderContext * context,
@@ -1002,6 +1073,8 @@ draw_key (GkbdKeyboardDrawingRenderContext * context,
 {
 	XkbShapeRec *shape;
 	GdkColor *color;
+	XkbOutlineRec *outline;
+	int origin_offset_x;
 	/* gint i; */
 
 	if (!drawing->xkb)
@@ -1030,9 +1103,9 @@ draw_key (GkbdKeyboardDrawingRenderContext * context,
 #endif
 
 	/* draw the primary outline */
-	draw_outline (context,
-		      shape->primary ? shape->primary : shape->outlines,
-		      color, key->angle, key->origin_x, key->origin_y);
+	outline = shape->primary ? shape->primary : shape->outlines;
+	draw_outline (context, outline, color, key->angle, key->origin_x,
+		      key->origin_y);
 #if 0
 	/* don't draw other outlines for now, since
 	 * the text placement does not take them into account 
@@ -1045,10 +1118,10 @@ draw_key (GkbdKeyboardDrawingRenderContext * context,
 			      key->angle, key->origin_x, key->origin_y);
 	}
 #endif
-
+	origin_offset_x = calc_shape_origin_offset_x (outline);
 	draw_key_label (context, drawing, key->keycode, key->angle,
-			key->origin_x, key->origin_y,
-			shape->bounds.x2, shape->bounds.y2);
+			key->origin_x + origin_offset_x, key->origin_y,
+			shape->bounds.x2, shape->bounds.y2, key->pressed);
 }
 
 static void
@@ -1057,6 +1130,7 @@ invalidate_region (GkbdKeyboardDrawing * drawing,
 		   gint origin_x, gint origin_y, XkbShapeRec * shape)
 {
 	GdkPoint points[4];
+	GtkAllocation alloc;
 	gint x_min, x_max, y_min, y_max;
 	gint x, y, width, height;
 	gint xx, yy;
@@ -1099,8 +1173,10 @@ invalidate_region (GkbdKeyboardDrawing * drawing,
 	    xkb_to_pixmap_coord (drawing->renderContext,
 				 y_max - y_min) + 12;
 
-	gtk_widget_queue_draw_area (GTK_WIDGET (drawing), x, y, width,
-				    height);
+	gtk_widget_get_allocation (GTK_WIDGET (drawing), &alloc);
+	gtk_widget_queue_draw_area (GTK_WIDGET (drawing),
+				    x + alloc.x, y + alloc.y,
+				    width, height);
 }
 
 static void
@@ -1116,9 +1192,8 @@ invalidate_indicator_doodad_region (GkbdKeyboardDrawing * drawing,
 			   doodad->doodad->indicator.left,
 			   doodad->origin_y +
 			   doodad->doodad->indicator.top,
-			   &drawing->xkb->geom->shapes[doodad->doodad->
-						       indicator.
-						       shape_ndx]);
+			   &drawing->xkb->geom->shapes[doodad->
+						       doodad->indicator.shape_ndx]);
 }
 
 static void
@@ -1132,8 +1207,8 @@ invalidate_key_region (GkbdKeyboardDrawing * drawing,
 			   key->angle,
 			   key->origin_x,
 			   key->origin_y,
-			   &drawing->xkb->geom->shapes[key->xkbkey->
-						       shape_ndx]);
+			   &drawing->xkb->geom->shapes[key->
+						       xkbkey->shape_ndx]);
 }
 
 static void
@@ -1152,7 +1227,7 @@ draw_text_doodad (GkbdKeyboardDrawingRenderContext * context,
 				 doodad->origin_y + text_doodad->top);
 
 	set_markup (context, text_doodad->text);
-	draw_pango_layout (context, drawing, doodad->angle, x, y);
+	draw_pango_layout (context, drawing, doodad->angle, x, y, FALSE);
 }
 
 static void
@@ -1314,13 +1389,13 @@ draw_keyboard_to_context (GkbdKeyboardDrawingRenderContext * context,
 }
 
 static gboolean
-create_cairo (GkbdKeyboardDrawing * drawing)
+prepare_cairo (GkbdKeyboardDrawing * drawing, cairo_t * cr)
 {
 	GtkStateType state;
-	if (drawing == NULL || drawing->surface == NULL)
+	if (drawing == NULL)
 		return FALSE;
-	drawing->renderContext->cr = cairo_create (drawing->surface);
 
+	drawing->renderContext->cr = cr;
 	state = gtk_widget_get_state (GTK_WIDGET (drawing));
 	drawing->renderContext->dark_color =
 	    &gtk_widget_get_style (GTK_WIDGET (drawing))->dark[state];
@@ -1328,15 +1403,7 @@ create_cairo (GkbdKeyboardDrawing * drawing)
 }
 
 static void
-destroy_cairo (GkbdKeyboardDrawing * drawing)
-{
-	cairo_destroy (drawing->renderContext->cr);
-	drawing->renderContext->cr = NULL;
-	drawing->renderContext->dark_color = NULL;
-}
-
-static void
-draw_keyboard (GkbdKeyboardDrawing * drawing)
+draw_keyboard (GkbdKeyboardDrawing * drawing, cairo_t * cr)
 {
 	GtkStateType state = gtk_widget_get_state (GTK_WIDGET (drawing));
 	GtkStyle *style = gtk_widget_get_style (GTK_WIDGET (drawing));
@@ -1347,18 +1414,24 @@ draw_keyboard (GkbdKeyboardDrawing * drawing)
 
 	gtk_widget_get_allocation (GTK_WIDGET (drawing), &allocation);
 
-	drawing->surface =
-	    gdk_window_create_similar_surface (gtk_widget_get_window (GTK_WIDGET (drawing)),
-			    CAIRO_CONTENT_COLOR, allocation.width, allocation.height);
-
-	if (create_cairo (drawing)) {
+	if (prepare_cairo (drawing, cr)) {
 		/* blank background */
-		gdk_cairo_set_source_color (drawing->renderContext->cr,
-					    &style->base[state]);
-		cairo_paint (drawing->renderContext->cr);
+		gdk_cairo_set_source_color (cr, &style->base[state]);
+		cairo_paint (cr);
+#ifdef KBDRAW_DEBUG
+		GdkColor red = { 0xFFFF, 0xFFFF, 0, 0 };
+		gdk_cairo_set_source_color (cr, &red);
+
+		cairo_move_to (cr, 0, 0);
+		cairo_line_to (cr, allocation.width, 0);
+		cairo_line_to (cr, allocation.width, allocation.height);
+		cairo_line_to (cr, 0, allocation.height);
+		cairo_close_path (cr);
+
+		cairo_stroke (cr);
+#endif
 
 		draw_keyboard_to_context (drawing->renderContext, drawing);
-		destroy_cairo (drawing);
 	}
 }
 
@@ -1395,41 +1468,12 @@ free_render_context (GkbdKeyboardDrawing * drawing)
 }
 
 static gboolean
-draw (GtkWidget * widget,
-      cairo_t *cr, GkbdKeyboardDrawing * drawing)
+draw (GtkWidget * widget, cairo_t * cr, GkbdKeyboardDrawing * drawing)
 {
-	GtkAllocation allocation;
-
 	if (!drawing->xkb)
 		return FALSE;
 
-	if (drawing->surface == NULL)
-		return FALSE;
-
-        cairo_set_source_surface (cr, drawing->surface, 0, 0);
-        cairo_paint (cr);
-
-	if (gtk_widget_has_focus (widget)) {
-		gtk_widget_get_allocation (widget, &allocation);
-		gtk_paint_focus (gtk_widget_get_style (widget),
-				 cr,
-				 gtk_widget_get_state (widget),
-				 widget, "keyboard-drawing",
-				 0, 0, allocation.width,
-				 allocation.height);
-	}
-
-	return FALSE;
-}
-
-static gboolean
-idle_redraw (gpointer user_data)
-{
-	GkbdKeyboardDrawing *drawing = user_data;
-
-	drawing->idle_redraw = 0;
-	draw_keyboard (drawing);
-	gtk_widget_queue_draw (GTK_WIDGET (drawing));
+	draw_keyboard (drawing, cr);
 	return FALSE;
 }
 
@@ -1477,18 +1521,12 @@ size_allocate (GtkWidget * widget,
 {
 	GkbdKeyboardDrawingRenderContext *context = drawing->renderContext;
 
-	if (drawing->surface) {
-                cairo_surface_destroy (drawing->surface);
-		drawing->surface = NULL;
-	}
-
 	if (!context_setup_scaling (context, drawing,
 				    allocation->width, allocation->height,
 				    50, 50))
 		return;
 
-	if (!drawing->idle_redraw)
-		drawing->idle_redraw = g_idle_add (idle_redraw, drawing);
+	gtk_widget_set_allocation (GTK_WIDGET (drawing), allocation);
 }
 
 static gint
@@ -1517,13 +1555,6 @@ key_event (GtkWidget * widget,
 
 	key->pressed = (event->type == GDK_KEY_PRESS);
 
-	if (create_cairo (drawing)) {
-		draw_key (drawing->renderContext, drawing, key);
-		redraw_overlapping_doodads (drawing->renderContext,
-					    drawing, key);
-		destroy_cairo (drawing);
-	}
-
 	invalidate_key_region (drawing, key);
 	return FALSE;
 }
@@ -1547,18 +1578,14 @@ unpress_keys (GkbdKeyboardDrawing * drawing)
 	if (!drawing->xkb)
 		return FALSE;
 
-	if (create_cairo (drawing)) {
-		for (i = drawing->xkb->min_key_code;
-		     i <= drawing->xkb->max_key_code; i++)
-			if (drawing->keys[i].pressed) {
-				drawing->keys[i].pressed = FALSE;
-				draw_key (drawing->renderContext, drawing,
-					  drawing->keys + i);
-				invalidate_key_region (drawing,
-						       drawing->keys + i);
-			}
-		destroy_cairo (drawing);
-	}
+	for (i = drawing->xkb->min_key_code;
+	     i <= drawing->xkb->max_key_code; i++)
+		if (drawing->keys[i].pressed) {
+			drawing->keys[i].pressed = FALSE;
+			draw_key (drawing->renderContext, drawing,
+				  drawing->keys + i);
+			invalidate_key_region (drawing, drawing->keys + i);
+		}
 
 	return FALSE;
 }
@@ -1694,8 +1721,8 @@ init_keys_and_doodads (GkbdKeyboardDrawing * drawing)
 				    drawing->xkb->geom->shapes +
 				    xkbkey->shape_ndx;
 				guint keycode = find_keycode (drawing,
-							      xkbkey->name.
-							      name);
+							      xkbkey->
+							      name.name);
 
 				if (keycode == INVALID_KEYCODE)
 					continue;
@@ -1806,8 +1833,9 @@ init_colors (GkbdKeyboardDrawing * drawing)
 
 	for (i = 0; i < drawing->xkb->geom->num_colors; i++) {
 		result =
-		    parse_xkb_color_spec (drawing->xkb->geom->colors[i].
-					  spec, drawing->colors + i);
+		    parse_xkb_color_spec (drawing->xkb->geom->
+					  colors[i].spec,
+					  drawing->colors + i);
 
 		if (!result)
 			g_warning
@@ -1878,14 +1906,6 @@ process_indicators_state_notify (XkbIndicatorNotifyEvent * iev,
 				&& drawing->physical_indicators[i]->on)) {
 				drawing->physical_indicators[i]->on =
 				    state;
-				if (create_cairo (drawing)) {
-					draw_doodad
-					    (drawing->renderContext,
-					     drawing,
-					     drawing->physical_indicators
-					     [i]);
-					destroy_cairo (drawing);
-				}
 				invalidate_indicator_doodad_region
 				    (drawing,
 				     drawing->physical_indicators[i]);
@@ -1937,8 +1957,8 @@ xkb_state_notify_event_filter (GdkXEvent * gdkxev,
 				process_indicators_state_notify (&
 								 ((XkbEvent
 								   *)
-								  gdkxev)->
-indicators, drawing);
+								  gdkxev)->indicators,
+drawing);
 			}
 			break;
 
@@ -1975,14 +1995,6 @@ destroy (GkbdKeyboardDrawing * drawing)
 		g_source_remove (drawing->timeout);
 		drawing->timeout = 0;
 	}
-	if (drawing->idle_redraw > 0) {
-		g_source_remove (drawing->idle_redraw);
-		drawing->idle_redraw = 0;
-	}
-
-        if (drawing->surface != NULL) {
-                cairo_surface_destroy (drawing->surface);
-        }
 }
 
 static void
@@ -1999,7 +2011,6 @@ gkbd_keyboard_drawing_init (GkbdKeyboardDrawing * drawing)
 
 	drawing->display =
 	    GDK_DISPLAY_XDISPLAY (gdk_display_get_default ());
-	printf ("dpy: %p\n", (void *) drawing->display);
 
 	if (!XkbQueryExtension
 	    (drawing->display, &opcode, &drawing->xkb_event_type, &error,
@@ -2007,9 +2018,6 @@ gkbd_keyboard_drawing_init (GkbdKeyboardDrawing * drawing)
 		g_critical
 		    ("XkbQueryExtension failed! Stuff probably won't work.");
 
-	printf ("evt/error/major/minor: %d/%d/%d/%d\n",
-		drawing->xkb_event_type, error, major, minor);
-
 	/* XXX: this stuff probably doesn't matter.. also, gdk_screen_get_default can fail */
 	if (gtk_widget_has_screen (GTK_WIDGET (drawing)))
 		drawing->screen_num =
@@ -2019,7 +2027,6 @@ gkbd_keyboard_drawing_init (GkbdKeyboardDrawing * drawing)
 		drawing->screen_num =
 		    gdk_screen_get_number (gdk_screen_get_default ());
 
-	drawing->surface = NULL;
 	alloc_render_context (drawing);
 
 	drawing->keyboard_items = NULL;
@@ -2029,6 +2036,7 @@ gkbd_keyboard_drawing_init (GkbdKeyboardDrawing * drawing)
 	drawing->track_config = 0;
 
 	gtk_widget_set_double_buffered (GTK_WIDGET (drawing), FALSE);
+	gtk_widget_set_has_window (GTK_WIDGET (drawing), FALSE);
 
 	/* XXX: XkbClientMapMask | XkbIndicatorMapMask | XkbNamesMask | XkbGeometryMask */
 	drawing->xkb = XkbGetKeyboard (drawing->display,
@@ -2111,10 +2119,42 @@ gkbd_keyboard_drawing_new (void)
 			(gkbd_keyboard_drawing_get_type (), NULL));
 }
 
+static GtkSizeRequestMode
+get_request_mode (GtkWidget * widget)
+{
+	return GTK_SIZE_REQUEST_HEIGHT_FOR_WIDTH;
+}
+
+static void
+get_preferred_width (GtkWidget * widget,
+		     gint * minimum_width, gint * natural_width)
+{
+	GdkScreen *scr = gdk_screen_get_default ();
+	gint w = gdk_screen_get_width (scr);
+	*minimum_width = *natural_width = w - (w >> 2);
+}
+
+static void
+get_preferred_height_for_width (GtkWidget * widget,
+				gint width,
+				gint * minimum_height,
+				gint * natural_height)
+{
+	GkbdKeyboardDrawing *drawing = GKBD_KEYBOARD_DRAWING (widget);
+	*minimum_height = *natural_height =
+	    width * drawing->xkb->geom->height_mm /
+	    drawing->xkb->geom->width_mm;
+}
+
 static void
 gkbd_keyboard_drawing_class_init (GkbdKeyboardDrawingClass * klass)
 {
 	klass->bad_keycode = NULL;
+	GTK_WIDGET_CLASS (klass)->get_preferred_height_for_width =
+	    get_preferred_height_for_width;
+	GTK_WIDGET_CLASS (klass)->get_preferred_width =
+	    get_preferred_width;
+	GTK_WIDGET_CLASS (klass)->get_request_mode = get_request_mode;
 
 	gkbd_keyboard_drawing_signals[BAD_KEYCODE] =
 	    g_signal_new ("bad-keycode", gkbd_keyboard_drawing_get_type (),
@@ -2457,19 +2497,85 @@ gkbd_keyboard_drawing_print (GkbdKeyboardDrawing * drawing,
 	g_object_unref (print);
 }
 
+void
+gkbd_keyboard_drawing_set_layout (GkbdKeyboardDrawing * drawing,
+				  const gchar * id)
+{
+	XklConfigRec *data;
+	char **p, *layout, *variant;
+	XkbComponentNamesRec component_names;
+	XklEngine *engine;
+
+	if (drawing == NULL)
+		return;
+
+	if (id == NULL) {
+		gkbd_keyboard_drawing_set_keyboard (drawing, NULL);
+		return;
+	}
+
+	engine = xkl_engine_get_instance (GDK_DISPLAY_XDISPLAY
+					  (gdk_display_get_default ()));
+
+	data = xkl_config_rec_new ();
+	if (xkl_config_rec_get_from_server (data, engine)) {
+		if ((p = data->layouts) != NULL)
+			g_strfreev (data->layouts);
+
+		if ((p = data->variants) != NULL)
+			g_strfreev (data->variants);
+
+		data->layouts = g_new0 (char *, 2);
+		data->variants = g_new0 (char *, 2);
+		if (gkbd_keyboard_config_split_items
+		    (id, &layout, &variant)
+		    && variant != NULL) {
+			data->layouts[0] =
+			    (layout == NULL) ? NULL : g_strdup (layout);
+			data->variants[0] =
+			    (variant == NULL) ? NULL : g_strdup (variant);
+		} else {
+			data->layouts[0] =
+			    (id == NULL) ? NULL : g_strdup (id);
+			data->variants[0] = NULL;
+		}
+
+		if (xkl_xkb_config_native_prepare
+		    (engine, data, &component_names)) {
+			if (!gkbd_keyboard_drawing_set_keyboard
+			    (drawing, &component_names))
+				gkbd_keyboard_drawing_set_keyboard
+				    (drawing, NULL);
+
+			xkl_xkb_config_native_cleanup
+			    (engine, &component_names);
+		} else {
+			xkl_debug (0, "Could not find the keyboard\n");
+		}
+	}
+	g_object_unref (G_OBJECT (data));
+}
+
 static void
-show_layout_response (GtkWidget * dialog, gint resp)
+gkbd_keyboard_drawing_dialog_set_layout_name (GtkWidget * dialog,
+					      const gchar * layout_name)
+{
+	char title[128] = "";
+	snprintf (title, sizeof (title), _("Keyboard Layout \"%s\""),
+		  layout_name);
+	gtk_window_set_title (GTK_WINDOW (dialog), title);
+	g_object_set_data_full (G_OBJECT (dialog), "layout_name",
+				g_strdup (layout_name), g_free);
+}
+
+static void
+gkbd_keyboard_drawing_dialog_response (GtkWidget * dialog, gint resp)
 {
 	GdkRectangle rect;
 	GtkWidget *kbdraw;
 	const gchar *groupName;
 
 	switch (resp) {
-	case GTK_RESPONSE_HELP:
-		gtk_show_uri (gtk_widget_get_screen (GTK_WIDGET (dialog)),
-			      "ghelp:gswitchit?layout-view",
-			      gtk_get_current_event_time (), NULL);
-		return;
 	case GTK_RESPONSE_CLOSE:
 		gtk_window_get_position (GTK_WINDOW (dialog), &rect.x,
 					 &rect.y);
@@ -2492,67 +2598,28 @@ show_layout_response (GtkWidget * dialog, gint resp)
 	}
 }
 
-GtkWidget *
-gkbd_keyboard_drawing_new_dialog (gint group, gchar * group_name)
+void
+gkbd_keyboard_drawing_dialog_set_group (GtkWidget * dialog,
+					XklConfigRegistry * registry,
+					gint group)
 {
-	static GkbdKeyboardDrawingGroupLevel groupsLevels[] = { {
-								 0, 1}, {
-									 0,
-									 3},
-	{
-	 0, 0}, {
-		 0, 2}
-	};
-	static GkbdKeyboardDrawingGroupLevel *pGroupsLevels[] = {
-		groupsLevels, groupsLevels + 1, groupsLevels + 2,
-		groupsLevels + 3
-	};
-
-	GtkBuilder *builder;
-	GtkWidget *dialog, *kbdraw;
 	XkbComponentNamesRec component_names;
 	XklConfigRec *xkl_data;
-	GdkRectangle *rect;
-	GError *error = NULL;
-	char title[128] = "";
 	XklEngine *engine =
 	    xkl_engine_get_instance (GDK_DISPLAY_XDISPLAY
 				     (gdk_display_get_default ()));
 
-	builder = gtk_builder_new ();
-	gtk_builder_add_from_file (builder, UIDIR "/show-layout.ui",
-				   &error);
-
-	if (error) {
-		g_error ("building ui from %s failed: %s",
-			 UIDIR "/show-layout.ui", error->message);
-		g_clear_error (&error);
-	}
-
-
-	dialog =
-	    GTK_WIDGET (gtk_builder_get_object
-			(builder, "gswitchit_layout_view"));
-	kbdraw = gkbd_keyboard_drawing_new ();
-
-	snprintf (title, sizeof (title), _("Keyboard Layout \"%s\""),
-		  group_name);
-	gtk_window_set_title (GTK_WINDOW (dialog), title);
-	g_object_set_data_full (G_OBJECT (dialog), "group_name",
-				g_strdup (group_name), g_free);
-
-	gkbd_keyboard_drawing_set_groups_levels (GKBD_KEYBOARD_DRAWING
-						 (kbdraw), pGroupsLevels);
-
 	xkl_data = xkl_config_rec_new ();
 	if (xkl_config_rec_get_from_server (xkl_data, engine)) {
 		int num_layouts = g_strv_length (xkl_data->layouts);
 		int num_variants = g_strv_length (xkl_data->variants);
 		if (group >= 0 && group < num_layouts
 		    && group < num_variants) {
-			char *l = g_strdup (xkl_data->layouts[group]);
-			char *v = g_strdup (xkl_data->variants[group]);
-			char **p;
+			XklConfigItem *xki = xkl_config_item_new ();
+			gchar *l = g_strdup (xkl_data->layouts[group]);
+			gchar *v = g_strdup (xkl_data->variants[group]);
+			const gchar *layout_name = NULL;
+			gchar **p;
 			int i;
 
 			if ((p = xkl_data->layouts) != NULL)
@@ -2573,37 +2640,82 @@ gkbd_keyboard_drawing_new_dialog (gint group, gchar * group_name)
 			xkl_data->variants[0] = v;
 			xkl_data->layouts[1] = xkl_data->variants[1] =
 			    NULL;
+
+			if (v[0] != 0) {
+				strncpy (xki->name, v,
+					 XKL_MAX_CI_NAME_LENGTH);
+				xki->name[XKL_MAX_CI_NAME_LENGTH - 1] = 0;
+				if (xkl_config_registry_find_variant
+				    (registry, l, xki))
+					layout_name = xki->description;
+			} else {
+				strncpy (xki->name, l,
+					 XKL_MAX_CI_NAME_LENGTH);
+				xki->name[XKL_MAX_CI_NAME_LENGTH - 1] = 0;
+				if (xkl_config_registry_find_layout
+				    (registry, xki))
+					layout_name = xki->description;
+			}
+			gkbd_keyboard_drawing_dialog_set_layout_name
+			    (dialog, layout_name);
+			g_object_unref (xki);
 		}
 
 		if (xkl_xkb_config_native_prepare
 		    (engine, xkl_data, &component_names)) {
-			gkbd_keyboard_drawing_set_keyboard
+			GtkWidget *kbdraw =
+			    g_object_get_data (G_OBJECT (dialog),
+					       "kbdraw");
+			if (!gkbd_keyboard_drawing_set_keyboard
 			    (GKBD_KEYBOARD_DRAWING (kbdraw),
-			     &component_names);
+			     &component_names))
+				gkbd_keyboard_drawing_set_keyboard
+				    (GKBD_KEYBOARD_DRAWING (kbdraw), NULL);
 			xkl_xkb_config_native_cleanup (engine,
 						       &component_names);
 		}
 	}
+
 	g_object_unref (G_OBJECT (xkl_data));
+}
+
+GtkWidget *
+gkbd_keyboard_drawing_dialog_new ()
+{
+	GtkBuilder *builder;
+	GtkWidget *dialog, *kbdraw;
+	GdkRectangle *rect;
+	GError *error = NULL;
+
+	builder = gtk_builder_new ();
+	gtk_builder_add_from_file (builder, UIDIR "/show-layout.ui",
+				   &error);
+
+	if (error) {
+		g_error ("building ui from %s failed: %s",
+			 UIDIR "/show-layout.ui", error->message);
+		g_clear_error (&error);
+	}
+
+	dialog =
+	    GTK_WIDGET (gtk_builder_get_object
+			(builder, "gswitchit_layout_view"));
+	kbdraw = gkbd_keyboard_drawing_new ();
+
+	gkbd_keyboard_drawing_set_groups_levels (GKBD_KEYBOARD_DRAWING
+						 (kbdraw), pGroupsLevels);
 
 	g_object_set_data (G_OBJECT (dialog), "builderData", builder);
 	g_signal_connect (G_OBJECT (dialog), "response",
-			  G_CALLBACK (show_layout_response), NULL);
-
-	rect = gkbd_preview_load_position ();
-	if (rect != NULL) {
-		gtk_window_move (GTK_WINDOW (dialog), rect->x, rect->y);
-		gtk_window_resize (GTK_WINDOW (dialog), rect->width,
-				   rect->height);
-		g_free (rect);
-	} else
-		gtk_window_resize (GTK_WINDOW (dialog), 700, 400);
+			  G_CALLBACK
+			  (gkbd_keyboard_drawing_dialog_response), NULL);
 
 	gtk_window_set_resizable (GTK_WINDOW (dialog), TRUE);
 
-	gtk_container_add (GTK_CONTAINER
-			   (gtk_builder_get_object
-			    (builder, "preview_vbox")), kbdraw);
+	gtk_box_pack_start (GTK_BOX
+			    (gtk_builder_get_object
+			     (builder, "preview_vbox")), kbdraw, TRUE,
+			    TRUE, 0);
 
 	g_object_set_data (G_OBJECT (dialog), "kbdraw", kbdraw);
 
@@ -2612,7 +2724,52 @@ gkbd_keyboard_drawing_new_dialog (gint group, gchar * group_name)
 				  g_object_get_data (G_OBJECT (dialog),
 						     "builderData"));
 
-	gtk_widget_show_all (dialog);
+	rect = gkbd_preview_load_position ();
+	if (rect != NULL) {
+		gtk_window_move (GTK_WINDOW (dialog), rect->x, rect->y);
+		g_free (rect);
+	}
 
 	return dialog;
 }
+
+void
+gkbd_keyboard_drawing_dialog_set_layout (GtkWidget * dialog,
+					 XklConfigRegistry * registry,
+					 const gchar * full_layout)
+{
+	const gchar *layout_name = "?";
+	XklConfigItem *xki = xkl_config_item_new ();
+	gchar *layout = NULL, *variant = NULL;
+
+	GkbdKeyboardDrawing *kbdraw =
+	    GKBD_KEYBOARD_DRAWING (g_object_get_data
+				   (G_OBJECT (dialog), "kbdraw"));
+
+	if (full_layout == NULL || full_layout[0] == 0)
+		return;
+
+	gkbd_keyboard_drawing_set_layout (kbdraw, full_layout);
+
+	if (gkbd_keyboard_config_split_items
+	    (full_layout, &layout, &variant)) {
+		if (variant != NULL) {
+			strncpy (xki->name, variant,
+				 XKL_MAX_CI_NAME_LENGTH);
+			xki->name[XKL_MAX_CI_NAME_LENGTH - 1] = 0;
+			if (xkl_config_registry_find_variant
+			    (registry, layout, xki))
+				layout_name = xki->description;
+		} else {
+			strncpy (xki->name, layout,
+				 XKL_MAX_CI_NAME_LENGTH);
+			xki->name[XKL_MAX_CI_NAME_LENGTH - 1] = 0;
+			if (xkl_config_registry_find_layout
+			    (registry, xki))
+				layout_name = xki->description;
+		}
+	}
+
+	gkbd_keyboard_drawing_dialog_set_layout_name (dialog, layout_name);
+	g_object_unref (xki);
+}
